# Copyright (c) 2015, Frappe Technologies Pvt. Ltd. and Contributors
# License: GNU General Public License v3. See license.txt

from __future__ import unicode_literals
import frappe
from frappe.model.document import Document
from frappe.model.mapper import get_mapped_doc
from frappe import _
from frappe.utils import flt
from erpnext.stock.doctype.quality_inspection_template.quality_inspection_template \
	import get_template_details

class QualityInspection(Document):
	def validate(self):
		if not self.readings and self.item_code:
			self.get_item_specification_details()

		if self.readings:
			self.set_status_based_on_acceptance_formula()

	def get_item_specification_details(self):
		if not self.quality_inspection_template:
			self.quality_inspection_template = frappe.db.get_value('Item',
				self.item_code, 'quality_inspection_template')

		if not self.quality_inspection_template: return

		self.set('readings', [])
		parameters = get_template_details(self.quality_inspection_template)
		for d in parameters:
			child = self.append('readings', {})
			child.specification = d.specification
			child.value = d.value
			child.acceptance_formula = d.acceptance_formula
			child.status = "Accepted"

	def get_quality_inspection_template(self):
		template = ''
		if self.bom_no:
			template = frappe.db.get_value('BOM', self.bom_no, 'quality_inspection_template')

		if not template:
			template = frappe.db.get_value('BOM', self.item_code, 'quality_inspection_template')

		self.quality_inspection_template = template
		self.get_item_specification_details()

	def on_submit(self):
		self.update_qc_reference()

	def on_cancel(self):
		self.update_qc_reference()

	def update_qc_reference(self):
		quality_inspection = self.name if self.docstatus == 1 else ""

		if self.reference_type == 'Job Card':
			if self.reference_name:
				frappe.db.sql("""
					UPDATE `tab{doctype}`
					SET quality_inspection = %s, modified = %s
					WHERE name = %s and production_item = %s
				""".format(doctype=self.reference_type),
					(quality_inspection, self.modified, self.reference_name, self.item_code))

		else:
			doctype = self.reference_type + ' Item'
			if self.reference_type == 'Stock Entry':
				doctype = 'Stock Entry Detail'

			if self.reference_type and self.reference_name:
				conditions = ""
				if self.batch_no:
					conditions += " and t1.batch_no = '%s'"%(self.batch_no)

				frappe.db.sql("""
					UPDATE
						`tab{child_doc}` t1, `tab{parent_doc}` t2
					SET
						t1.quality_inspection = %s, t2.modified = %s
					WHERE
						t1.parent = %s
						and t1.item_code = %s
						and t1.parent = t2.name
						{conditions}
				""".format(parent_doc=self.reference_type, child_doc=doctype, conditions=conditions),
					(quality_inspection, self.modified, self.reference_name, self.item_code))

	def set_status_based_on_acceptance_formula(self):
		for reading in self.readings:
			if not reading.acceptance_formula: continue

			condition = reading.acceptance_formula
			data = {}
			for i in range(1, 11):
				field = "reading_" + str(i)
				data[field] = flt(reading.get(field)) or 0

			try:
				result = frappe.safe_eval(condition, None, data)
				reading.status = "Accepted" if result else "Rejected"
			except SyntaxError:
				frappe.throw(_("Row #{0}: Acceptance Criteria Formula is incorrect.").format(reading.idx),
					title=_("Invalid Formula"))
			except NameError as e:
				field = frappe.bold(e.args[0].split()[1])
				frappe.throw(_("Row #{0}: {1} is not a valid reading field. Please refer to the field description.")
					.format(reading.idx, field),
					title=_("Invalid Formula"))


@frappe.whitelist()
@frappe.validate_and_sanitize_search_inputs
def item_query(doctype, txt, searchfield, start, page_len, filters):
	if filters.get("from"):
		from frappe.desk.reportview import get_match_cond
		mcond = get_match_cond(filters["from"])
		cond, qi_condition = "", "and (quality_inspection is null or quality_inspection = '')"

		if filters.get("parent"):
			if filters.get('from') in ['Purchase Invoice Item', 'Purchase Receipt Item']\
					and filters.get("inspection_type") != "In Process":
				cond = """and item_code in (select name from `tabItem` where
					inspection_required_before_purchase = 1)"""
			elif filters.get('from') in ['Sales Invoice Item', 'Delivery Note Item']\
					and filters.get("inspection_type") != "In Process":
				cond = """and item_code in (select name from `tabItem` where
					inspection_required_before_delivery = 1)"""
			elif filters.get('from') == 'Stock Entry Detail':
				cond = """and s_warehouse is null"""

			if filters.get('from') in ['Supplier Quotation Item']:
				qi_condition = ""

			return frappe.db.sql("""
					SELECT item_code
					FROM `tab{doc}`
					WHERE parent=%(parent)s and docstatus < 2 and item_code like %(txt)s
					{qi_condition} {cond} {mcond}
					ORDER BY item_code limit {start}, {page_len}
				""".format(doc=filters.get('from'),
					cond = cond, mcond = mcond, start = start,
					page_len = page_len, qi_condition = qi_condition),
					{'parent': filters.get('parent'), 'txt': "%%%s%%" % txt})

		elif filters.get("reference_name"):
			return frappe.db.sql("""
					SELECT production_item
					FROM `tab{doc}`
					WHERE name = %(reference_name)s and docstatus < 2 and production_item like %(txt)s
					{qi_condition} {cond} {mcond}
					ORDER BY production_item
					LIMIT {start}, {page_len}
				""".format(doc=filters.get("from"),
					cond = cond, mcond = mcond, start = start,
					page_len = page_len, qi_condition = qi_condition),
					{'reference_name': filters.get('reference_name'), 'txt': "%%%s%%" % txt})

@frappe.whitelist()
@frappe.validate_and_sanitize_search_inputs
def quality_inspection_query(doctype, txt, searchfield, start, page_len, filters):
	return frappe.get_all('Quality Inspection',
		limit_start=start,
		limit_page_length=page_len,
		filters = {
			'docstatus': 1,
			'name': ('like', '%%%s%%' % txt),
			'item_code': filters.get("item_code"),
			'reference_name': ('in', [filters.get("reference_name", ''), ''])
		}, as_list=1)

@frappe.whitelist()
def make_quality_inspection(source_name, target_doc=None):
	def postprocess(source, doc):
		doc.inspected_by = frappe.session.user
		doc.get_quality_inspection_template()

	doc = get_mapped_doc("BOM", source_name, {
		'BOM': {
			"doctype": "Quality Inspection",
			"validation": {
				"docstatus": ["=", 1]
			},
			"field_map": {
				"name": "bom_no",
				"item": "item_code",
				"stock_uom": "uom",
				"stock_qty": "qty"
			},
		}
	}, target_doc, postprocess)

	return doc
